@@ -18,6 +18,7 @@ use sixtyfps_corelib as corelib;
 use corelib::graphics::Point;
 use corelib::input::{InternalKeyCode, KeyEvent, KeyEventType, KeyboardModifiers, MouseEvent};
 use corelib::window::*;
+use corelib::SharedString;
 use std::cell::RefCell;
 use std::rc::{Rc, Weak};
 
@@ -225,8 +226,8 @@ pub fn run(quit_behavior: sixtyfps_corelib::backend::EventLoopQuitBehavior) {
                 winit::event::Event::RedrawRequested(id) => {
                     corelib::animations::update_animations();
                     ALL_WINDOWS.with(|windows| {
-                        if let Some(Some(window)) =
-                            windows.borrow().get(&id).map(|weakref| weakref.upgrade())
+                        if let Some(window) =
+                            windows.borrow().get(&id).and_then(|weakref| weakref.upgrade())
                         {
                             window.draw();
                         }
@@ -237,10 +238,9 @@ pub fn run(quit_behavior: sixtyfps_corelib::backend::EventLoopQuitBehavior) {
                     window_id,
                 } => {
                     ALL_WINDOWS.with(|windows| {
-                        if let Some(Some(window)) =
-                            windows.borrow().get(&window_id).map(|weakref| weakref.upgrade())
+                        if let Some(window) =
+                            windows.borrow().get(&window_id).and_then(|weakref| weakref.upgrade())
                         {
-                            window.refresh_window_scale_factor();
                             let size = size.to_logical(window.scale_factor() as f64);
                             window.set_geometry(size.width, size.height);
                         }
@@ -255,12 +255,13 @@ pub fn run(quit_behavior: sixtyfps_corelib::backend::EventLoopQuitBehavior) {
                     window_id,
                 } => {
                     ALL_WINDOWS.with(|windows| {
-                        if let Some(Some(window)) =
-                            windows.borrow().get(&window_id).map(|weakref| weakref.upgrade())
+                        if let Some(window) =
+                            windows.borrow().get(&window_id).and_then(|weakref| weakref.upgrade())
                         {
                             let size = size.to_logical(scale_factor);
                             window.set_geometry(size.width, size.height);
-                            window.set_scale_factor(scale_factor as f32);
+                            let runtime_window = window.self_weak.upgrade().unwrap();
+                            runtime_window.set_scale_factor(scale_factor as f32);
                         }
                     });
                 }
@@ -272,8 +273,8 @@ pub fn run(quit_behavior: sixtyfps_corelib::backend::EventLoopQuitBehavior) {
                 } => {
                     corelib::animations::update_animations();
                     ALL_WINDOWS.with(|windows| {
-                        if let Some(Some(window)) =
-                            windows.borrow().get(window_id).map(|weakref| weakref.upgrade())
+                        if let Some(window) =
+                            windows.borrow().get(window_id).and_then(|weakref| weakref.upgrade())
                         {
                             let ev = match state {
                                 winit::event::ElementState::Pressed => {
@@ -296,8 +297,8 @@ pub fn run(quit_behavior: sixtyfps_corelib::backend::EventLoopQuitBehavior) {
                 } => {
                     corelib::animations::update_animations();
                     ALL_WINDOWS.with(|windows| {
-                        if let Some(Some(window)) =
-                            windows.borrow().get(window_id).map(|weakref| weakref.upgrade())
+                        if let Some(window) =
+                            windows.borrow().get(window_id).and_then(|weakref| weakref.upgrade())
                         {
                             let location = touch.location.to_logical(window.scale_factor() as f64);
                             let pos = euclid::point2(location.x, location.y);
@@ -324,8 +325,8 @@ pub fn run(quit_behavior: sixtyfps_corelib::backend::EventLoopQuitBehavior) {
                 } => {
                     corelib::animations::update_animations();
                     ALL_WINDOWS.with(|windows| {
-                        if let Some(Some(window)) =
-                            windows.borrow().get(&window_id).map(|weakref| weakref.upgrade())
+                        if let Some(window) =
+                            windows.borrow().get(&window_id).and_then(|weakref| weakref.upgrade())
                         {
                             let position = position.to_logical(window.scale_factor() as f64);
                             cursor_pos = euclid::point2(position.x, position.y);
@@ -343,8 +344,10 @@ pub fn run(quit_behavior: sixtyfps_corelib::backend::EventLoopQuitBehavior) {
                     if pressed {
                         corelib::animations::update_animations();
                         ALL_WINDOWS.with(|windows| {
-                            if let Some(Some(window)) =
-                                windows.borrow().get(&window_id).map(|weakref| weakref.upgrade())
+                            if let Some(window) = windows
+                                .borrow()
+                                .get(&window_id)
+                                .and_then(|weakref| weakref.upgrade())
                             {
                                 pressed = false;
                                 window.process_mouse_input(MouseEvent::MouseExit);
@@ -359,8 +362,8 @@ pub fn run(quit_behavior: sixtyfps_corelib::backend::EventLoopQuitBehavior) {
                 } => {
                     corelib::animations::update_animations();
                     ALL_WINDOWS.with(|windows| {
-                        if let Some(Some(window)) =
-                            windows.borrow().get(window_id).map(|weakref| weakref.upgrade())
+                        if let Some(window) =
+                            windows.borrow().get(window_id).and_then(|weakref| weakref.upgrade())
                         {
                             let delta = match delta {
                                 winit::event::MouseScrollDelta::LineDelta(lx, ly) => {
@@ -384,9 +387,15 @@ pub fn run(quit_behavior: sixtyfps_corelib::backend::EventLoopQuitBehavior) {
                 } => {
                     corelib::animations::update_animations();
                     ALL_WINDOWS.with(|windows| {
-                        if let Some(Some(window)) =
-                            windows.borrow().get(window_id).map(|weakref| weakref.upgrade())
+                        if let Some(window) =
+                            windows.borrow().get(window_id).and_then(|weakref| weakref.upgrade())
                         {
+                            window.currently_pressed_key_code.set(match input.state {
+                                winit::event::ElementState::Pressed => {
+                                    input.virtual_keycode.clone()
+                                }
+                                _ => None,
+                            });
                             if let Some(key_code) =
                                 input.virtual_keycode.and_then(|virtual_keycode| {
                                     match virtual_keycode {
@@ -440,35 +449,46 @@ pub fn run(quit_behavior: sixtyfps_corelib::backend::EventLoopQuitBehavior) {
                     ref window_id,
                     event: winit::event::WindowEvent::ReceivedCharacter(ch),
                 } => {
-                    if !ch.is_control() {
-                        corelib::animations::update_animations();
-                        ALL_WINDOWS.with(|windows| {
-                            if let Some(Some(window)) =
-                                windows.borrow().get(window_id).map(|weakref| weakref.upgrade())
-                            {
-                                let modifiers = window.current_keyboard_modifiers();
+                    corelib::animations::update_animations();
+                    ALL_WINDOWS.with(|windows| {
+                        if let Some(window) =
+                            windows.borrow().get(window_id).and_then(|weakref| weakref.upgrade())
+                        {
+                            // On Windows, X11 and Wayland sequences like Ctrl+C will send a ReceivedCharacter after the pressed keyboard input event,
+                            // with a control character. We choose not to forward those but try to use the current key code instead.
+                            let text: Option<SharedString> = if ch.is_control() {
+                                window
+                                    .currently_pressed_key_code
+                                    .take()
+                                    .and_then(winit_key_code_to_string)
+                            } else {
+                                Some(ch.to_string().into())
+                            };
 
-                                let mut event = KeyEvent {
-                                    event_type: KeyEventType::KeyPressed,
-                                    text: ch.to_string().into(),
-                                    modifiers,
-                                };
+                            let text = match text {
+                                Some(text) => text,
+                                None => return,
+                            };
 
-                                window.self_weak.upgrade().unwrap().process_key_input(&event);
+                            let modifiers = window.current_keyboard_modifiers();
 
-                                event.event_type = KeyEventType::KeyReleased;
-                                window.self_weak.upgrade().unwrap().process_key_input(&event);
-                            }
-                        });
-                    }
+                            let mut event =
+                                KeyEvent { event_type: KeyEventType::KeyPressed, text, modifiers };
+
+                            window.self_weak.upgrade().unwrap().process_key_input(&event);
+
+                            event.event_type = KeyEventType::KeyReleased;
+                            window.self_weak.upgrade().unwrap().process_key_input(&event);
+                        }
+                    });
                 }
                 winit::event::Event::WindowEvent {
                     ref window_id,
                     event: winit::event::WindowEvent::ModifiersChanged(state),
                 } => {
                     ALL_WINDOWS.with(|windows| {
-                        if let Some(Some(window)) =
-                            windows.borrow().get(window_id).map(|weakref| weakref.upgrade())
+                        if let Some(window) =
+                            windows.borrow().get(window_id).and_then(|weakref| weakref.upgrade())
                         {
                             // To provide an easier cross-platform behavior, we map the command key to control
                             // on macOS, and control to meta.
@@ -492,8 +512,8 @@ pub fn run(quit_behavior: sixtyfps_corelib::backend::EventLoopQuitBehavior) {
                     event: winit::event::WindowEvent::Focused(have_focus),
                 } => {
                     ALL_WINDOWS.with(|windows| {
-                        if let Some(Some(window)) =
-                            windows.borrow().get(window_id).map(|weakref| weakref.upgrade())
+                        if let Some(window) =
+                            windows.borrow().get(window_id).and_then(|weakref| weakref.upgrade())
                         {
                             window.self_weak.upgrade().unwrap().set_focus(have_focus);
                         }
@@ -502,8 +522,8 @@ pub fn run(quit_behavior: sixtyfps_corelib::backend::EventLoopQuitBehavior) {
 
                 winit::event::Event::UserEvent(CustomEvent::UpdateWindowProperties(window_id)) => {
                     ALL_WINDOWS.with(|windows| {
-                        if let Some(Some(window)) =
-                            windows.borrow().get(&window_id).map(|weakref| weakref.upgrade())
+                        if let Some(window) =
+                            windows.borrow().get(&window_id).and_then(|weakref| weakref.upgrade())
                         {
                             window.self_weak.upgrade().unwrap().update_window_properties();
                         }
@@ -570,3 +590,68 @@ pub fn run(quit_behavior: sixtyfps_corelib::backend::EventLoopQuitBehavior) {
         winit_loop.run(run_fn)
     }
 }
+
+// This function is called when we receive a control character via WindowEvent::ReceivedCharacter and
+// instead want to use the last virtual key code. That happens when for example pressing Ctrl+some_key
+// on Windows/X11/Wayland. This function may be missing mappings, it's trying to cover what we may be
+// getting when we're getting control character sequences.
+fn winit_key_code_to_string(virtual_keycode: winit::event::VirtualKeyCode) -> Option<SharedString> {
+    use winit::event::VirtualKeyCode;
+    Some(
+        match virtual_keycode {
+            VirtualKeyCode::Key1 => "1",
+            VirtualKeyCode::Key2 => "2",
+            VirtualKeyCode::Key3 => "3",
+            VirtualKeyCode::Key4 => "4",
+            VirtualKeyCode::Key5 => "5",
+            VirtualKeyCode::Key6 => "6",
+            VirtualKeyCode::Key7 => "7",
+            VirtualKeyCode::Key8 => "8",
+            VirtualKeyCode::Key9 => "9",
+            VirtualKeyCode::Key0 => "0",
+            VirtualKeyCode::A => "a",
+            VirtualKeyCode::B => "b",
+            VirtualKeyCode::C => "c",
+            VirtualKeyCode::D => "d",
+            VirtualKeyCode::E => "e",
+            VirtualKeyCode::F => "f",
+            VirtualKeyCode::G => "g",
+            VirtualKeyCode::H => "h",
+            VirtualKeyCode::I => "i",
+            VirtualKeyCode::J => "j",
+            VirtualKeyCode::K => "k",
+            VirtualKeyCode::L => "l",
+            VirtualKeyCode::M => "m",
+            VirtualKeyCode::N => "n",
+            VirtualKeyCode::O => "o",
+            VirtualKeyCode::P => "p",
+            VirtualKeyCode::Q => "q",
+            VirtualKeyCode::R => "r",
+            VirtualKeyCode::S => "s",
+            VirtualKeyCode::T => "t",
+            VirtualKeyCode::U => "u",
+            VirtualKeyCode::V => "v",
+            VirtualKeyCode::W => "w",
+            VirtualKeyCode::X => "x",
+            VirtualKeyCode::Y => "y",
+            VirtualKeyCode::Z => "z",
+            VirtualKeyCode::Space => " ",
+            VirtualKeyCode::Caret => "^",
+            VirtualKeyCode::Apostrophe => "'",
+            VirtualKeyCode::Asterisk => "*",
+            VirtualKeyCode::Backslash => "\\",
+            VirtualKeyCode::Colon => ":",
+            VirtualKeyCode::Comma => ",",
+            VirtualKeyCode::Equals => "=",
+            VirtualKeyCode::Grave => "`",
+            VirtualKeyCode::Minus => "-",
+            VirtualKeyCode::Period => ".",
+            VirtualKeyCode::Plus => "+",
+            VirtualKeyCode::Semicolon => ";",
+            VirtualKeyCode::Slash => "/",
+            VirtualKeyCode::Tab => "\t",
+            _ => return None,
+        }
+        .into(),
+    )
+}
