@@ -12,7 +12,10 @@ LICENSE END */
 use femtovg::TextContext;
 #[cfg(target_os = "windows")]
 use font_kit::loader::Loader;
-use sixtyfps_corelib::graphics::{FontMetrics as FontMetricsTrait, FontRequest, Size};
+use sixtyfps_corelib::graphics::{FontRequest, Point, Size};
+use sixtyfps_corelib::items::{
+    TextHorizontalAlignment, TextOverflow, TextVerticalAlignment, TextWrap,
+};
 use sixtyfps_corelib::{SharedString, SharedVector};
 #[cfg(target_arch = "wasm32")]
 use std::cell::Cell;
@@ -251,22 +254,6 @@ pub struct Font {
 }
 
 impl Font {
-    pub fn measure(&self, letter_spacing: f32, text: &str) -> femtovg::TextMetrics {
-        self.text_context
-            .measure_text(0., 0., text, self.init_paint(letter_spacing, femtovg::Paint::default()))
-            .unwrap()
-    }
-
-    pub fn height(&self) -> f32 {
-        self.text_context
-            .measure_font(self.init_paint(
-                /*letter spacing does not influence height*/ 0.,
-                femtovg::Paint::default(),
-            ))
-            .unwrap()
-            .height()
-    }
-
     pub fn init_paint(&self, letter_spacing: f32, mut paint: femtovg::Paint) -> femtovg::Paint {
         paint.set_font(&self.fonts);
         paint.set_font_size(self.pixel_size);
@@ -305,66 +292,30 @@ impl Font {
     }
 }
 
-pub struct FontMetrics {
-    font: Font,
-    letter_spacing: Option<f32>,
-    scale_factor: f32,
-}
-
-impl FontMetrics {
-    pub(crate) fn new(
-        graphics_cache: &mut ItemGraphicsCache,
-        item_graphics_cache_data: &sixtyfps_corelib::item_rendering::CachedRenderingData,
-        font_request_fn: impl Fn() -> sixtyfps_corelib::graphics::FontRequest,
-        scale_factor: core::pin::Pin<&sixtyfps_corelib::Property<f32>>,
-        reference_text: core::pin::Pin<&sixtyfps_corelib::Property<SharedString>>,
-    ) -> Self {
-        let font = graphics_cache
-            .load_item_graphics_cache_with_function(item_graphics_cache_data, || {
-                Some(super::ItemGraphicsCacheEntry::Font(FONT_CACHE.with(|cache| {
-                    cache.borrow_mut().font(
-                        font_request_fn(),
-                        scale_factor.get(),
-                        &reference_text.get(),
-                    )
-                })))
-            })
-            .unwrap()
-            .as_font()
-            .clone();
-        Self {
-            font,
-            letter_spacing: font_request_fn().letter_spacing,
-            scale_factor: scale_factor.get(),
-        }
-    }
-}
-
-impl FontMetricsTrait for FontMetrics {
-    fn text_size(&self, text: &str, max_width: Option<f32>) -> Size {
-        self.font.text_size(
-            self.letter_spacing.unwrap_or_default(),
-            text,
-            max_width.map(|x| x * self.scale_factor),
-        ) / self.scale_factor
-    }
-
-    fn line_height(&self) -> f32 {
-        self.font.height()
-    }
-
-    fn text_offset_for_x_position(&self, text: &str, x: f32) -> usize {
-        let x = x * self.scale_factor;
-        let metrics = self.font.measure(self.letter_spacing.unwrap_or_default(), text);
-        let mut current_x = 0.;
-        for glyph in metrics.glyphs {
-            if current_x + glyph.advance_x / 2. >= x {
-                return glyph.byte_index;
-            }
-            current_x += glyph.advance_x;
-        }
-        text.len()
-    }
+pub(crate) fn text_size(
+    graphics_cache: &RefCell<ItemGraphicsCache>,
+    item_graphics_cache: &sixtyfps_corelib::item_rendering::CachedRenderingData,
+    font_request_fn: impl Fn() -> sixtyfps_corelib::graphics::FontRequest,
+    scale_factor: std::pin::Pin<&sixtyfps_corelib::Property<f32>>,
+    text: &str,
+    max_width: Option<f32>,
+) -> Size {
+    let cached_font = item_graphics_cache
+        .get_or_update(graphics_cache, || {
+            Some(super::ItemGraphicsCacheEntry::Font(FONT_CACHE.with(|cache| {
+                cache.borrow_mut().font(
+                    font_request_fn(),
+                    scale_factor.get(),
+                    // FIXME: there is no dependency to the text property
+                    text,
+                )
+            })))
+        })
+        .unwrap();
+    let font = cached_font.as_font();
+    let letter_spacing = font_request_fn().letter_spacing.unwrap_or_default();
+    let scale_factor = scale_factor.get();
+    font.text_size(letter_spacing, text, max_width.map(|x| x * scale_factor)) / scale_factor
 }
 
 pub struct FontCache {
@@ -420,3 +371,118 @@ impl FontCache {
         }
     }
 }
+
+/// Layout the given string in lines, and call the `layout_line` callback with the line to draw at position y.
+/// The signature of the `layout_line` function is: `(canvas, text, pos, start_index, line_metrics)`.
+/// start index is the starting byte of the text in the string.
+/// Returns the baseline y coordinate.
+pub(crate) fn layout_text_lines(
+    string: &str,
+    font: &Font,
+    Size { width: max_width, height: max_height, .. }: Size,
+    (horizontal_alignment, vertical_alignment): (TextHorizontalAlignment, TextVerticalAlignment),
+    wrap: TextWrap,
+    overflow: TextOverflow,
+    single_line: bool,
+    paint: femtovg::Paint,
+    mut layout_line: impl FnMut(&str, Point, usize, &femtovg::TextMetrics),
+) -> f32 {
+    let wrap = wrap == TextWrap::word_wrap;
+    let elide = overflow == TextOverflow::elide;
+
+    let text_context = FONT_CACHE.with(|cache| cache.borrow().text_context.clone());
+    let font_metrics = text_context.measure_font(paint).unwrap();
+    let font_height = font_metrics.height();
+
+    let text_height = || {
+        if single_line {
+            font_height
+        } else {
+            // Note: this is kind of doing twice the layout because text_size also does it
+            font.text_size(
+                paint.letter_spacing(),
+                string,
+                if wrap { Some(max_width) } else { None },
+            )
+            .height
+        }
+    };
+
+    let mut process_line =
+        |text: &str, y: f32, start: usize, line_metrics: &femtovg::TextMetrics| {
+            let x = match horizontal_alignment {
+                TextHorizontalAlignment::left => 0.,
+                TextHorizontalAlignment::center => max_width / 2. - line_metrics.width() / 2.,
+                TextHorizontalAlignment::right => max_width - line_metrics.width(),
+            };
+            layout_line(text, Point::new(x, y), start, line_metrics);
+        };
+
+    let baseline_y = match vertical_alignment {
+        TextVerticalAlignment::top => 0.,
+        TextVerticalAlignment::center => max_height / 2. - text_height() / 2.,
+        TextVerticalAlignment::bottom => max_height - text_height(),
+    };
+    let mut y = baseline_y;
+    let mut start = 0;
+    'lines: while start < string.len() && y + font_height <= max_height {
+        if wrap && (!elide || y + 2. * font_height <= max_height) {
+            let index = text_context.break_text(max_width, &string[start..], paint).unwrap();
+            if index == 0 {
+                // FIXME the word is too big to be shown, but we should still break, ideally
+                break;
+            }
+            let index = start + index;
+            let line = &string[start..index];
+            let text_metrics = text_context.measure_text(0., 0., line, paint).unwrap();
+            process_line(line, y, start, &text_metrics);
+            y += font_height;
+            start = index;
+        } else {
+            let index = if single_line {
+                string.len()
+            } else {
+                string[start..].find('\n').map_or(string.len(), |i| start + i + 1)
+            };
+            let line = &string[start..index];
+            let text_metrics = text_context.measure_text(0., 0., line, paint).unwrap();
+            let elide_last_line =
+                elide && index < string.len() && y + 2. * font_height > max_height;
+            if text_metrics.width() > max_width || elide_last_line {
+                let w = max_width
+                    - if elide {
+                        text_context.measure_text(0., 0., "…", paint).unwrap().width()
+                    } else {
+                        0.
+                    };
+                let mut current_x = 0.;
+                for glyph in &text_metrics.glyphs {
+                    current_x += glyph.advance_x;
+                    if current_x >= w {
+                        let txt = &line[..glyph.byte_index];
+                        if elide {
+                            let elided = format!("{}…", txt);
+                            process_line(&elided, y, start, &text_metrics);
+                        } else {
+                            process_line(txt, y, start, &text_metrics);
+                        }
+                        y += font_height;
+                        start = index;
+                        continue 'lines;
+                    }
+                }
+                if elide_last_line {
+                    let elided = format!("{}…", line);
+                    process_line(&elided, y, start, &text_metrics);
+                    y += font_height;
+                    start = index;
+                    continue 'lines;
+                }
+            }
+            process_line(line, y, start, &text_metrics);
+            y += font_height;
+            start = index;
+        }
+    }
+    baseline_y
+}
