@@ -15,7 +15,7 @@ use std::rc::Rc;
 use sixtyfps_corelib::graphics::{SharedImageBuffer, Size};
 #[cfg(target_arch = "wasm32")]
 use sixtyfps_corelib::Property;
-use sixtyfps_corelib::{slice::Slice, ImageInner, SharedString};
+use sixtyfps_corelib::{items::ImageRendering, slice::Slice, ImageInner, SharedString};
 
 use super::{CanvasRc, GLItemRenderer};
 
@@ -236,21 +236,28 @@ impl CachedImage {
     // Upload the image to the GPU? if that hasn't happened yet. This function could take just a canvas
     // as parameter, but since an upload requires a current context, this is "enforced" by taking
     // a renderer instead (which implies a current context).
-    pub fn ensure_uploaded_to_gpu(&self, current_renderer: &GLItemRenderer) -> femtovg::ImageId {
+    pub fn ensure_uploaded_to_gpu(
+        &self,
+        current_renderer: &GLItemRenderer,
+        scaling: Option<ImageRendering>,
+    ) -> femtovg::ImageId {
         use std::convert::TryFrom;
 
-        let canvas = &current_renderer.shared_data.canvas;
+        let canvas = &current_renderer.canvas;
+
+        let image_flags = match scaling.unwrap_or_default() {
+            ImageRendering::smooth => femtovg::ImageFlags::empty(),
+            ImageRendering::pixelated => femtovg::ImageFlags::NEAREST,
+        };
 
         let img = &mut *self.0.borrow_mut();
         if let ImageData::DecodedImage(decoded_image) = img {
             let image_id = match femtovg::ImageSource::try_from(&*decoded_image) {
-                Ok(image_source) => {
-                    canvas.borrow_mut().create_image(image_source, femtovg::ImageFlags::empty())
-                }
+                Ok(image_source) => canvas.borrow_mut().create_image(image_source, image_flags),
                 Err(_) => {
                     let converted = image::DynamicImage::ImageRgba8(decoded_image.to_rgba8());
                     let image_source = femtovg::ImageSource::try_from(&converted).unwrap();
-                    canvas.borrow_mut().create_image(image_source, femtovg::ImageFlags::empty())
+                    canvas.borrow_mut().create_image(image_source, image_flags)
                 }
             }
             .unwrap();
@@ -258,16 +265,15 @@ impl CachedImage {
             *img = Texture { id: image_id, canvas: canvas.clone() }.into()
         } else if let ImageData::EmbeddedImage(buffer) = img {
             let (image_source, flags) = image_buffer_to_image_source(buffer);
-            let image_id = canvas.borrow_mut().create_image(image_source, flags).unwrap();
+            let image_id =
+                canvas.borrow_mut().create_image(image_source, flags | image_flags).unwrap();
             *img = Texture { id: image_id, canvas: canvas.clone() }.into()
         }
 
         #[cfg(target_arch = "wasm32")]
         if let ImageData::HTMLImage(html_image) = img {
-            let image_id = canvas
-                .borrow_mut()
-                .create_image(&html_image.dom_element, femtovg::ImageFlags::empty())
-                .unwrap();
+            let image_id =
+                canvas.borrow_mut().create_image(&html_image.dom_element, image_flags).unwrap();
             *img = Texture { id: image_id, canvas: canvas.clone() }.into()
         }
 
@@ -284,22 +290,26 @@ impl CachedImage {
         &self,
         current_renderer: &GLItemRenderer,
         target_size: euclid::default::Size2D<u32>,
+        scaling: ImageRendering,
     ) -> Option<Self> {
         use std::convert::TryFrom;
 
-        let canvas = &current_renderer.shared_data.canvas;
+        let canvas = &current_renderer.canvas;
+
+        let image_flags = match scaling {
+            ImageRendering::smooth => femtovg::ImageFlags::empty(),
+            ImageRendering::pixelated => femtovg::ImageFlags::NEAREST,
+        };
 
         match &*self.0.borrow() {
             ImageData::Texture(_) => None, // internal error: Cannot call upload_to_gpu on previously uploaded image,
             ImageData::DecodedImage(decoded_image) => {
                 let image_id = match femtovg::ImageSource::try_from(&*decoded_image) {
-                    Ok(image_source) => {
-                        canvas.borrow_mut().create_image(image_source, femtovg::ImageFlags::empty())
-                    }
+                    Ok(image_source) => canvas.borrow_mut().create_image(image_source, image_flags),
                     Err(_) => {
                         let converted = image::DynamicImage::ImageRgba8(decoded_image.to_rgba8());
                         let image_source = femtovg::ImageSource::try_from(&converted).unwrap();
-                        canvas.borrow_mut().create_image(image_source, femtovg::ImageFlags::empty())
+                        canvas.borrow_mut().create_image(image_source, image_flags)
                     }
                 }
                 .unwrap();
@@ -322,10 +332,8 @@ impl CachedImage {
             },
             #[cfg(target_arch = "wasm32")]
             ImageData::HTMLImage(html_image) => html_image.size().map(|_| {
-                let image_id = canvas
-                    .borrow_mut()
-                    .create_image(&html_image.dom_element, femtovg::ImageFlags::empty())
-                    .unwrap();
+                let image_id =
+                    canvas.borrow_mut().create_image(&html_image.dom_element, image_flags).unwrap();
                 Self::new_on_gpu(canvas, image_id)
             }),
         }
@@ -415,25 +423,35 @@ impl CachedImage {
     }
 }
 
-#[derive(PartialEq, Eq, Hash, Debug)]
-pub enum ImageCacheKey {
+#[derive(PartialEq, Eq, Hash, Debug, derive_more::From)]
+enum CachedImageSourceKey {
     Path(String),
     EmbeddedData(by_address::ByAddress<&'static [u8]>),
 }
 
+#[derive(PartialEq, Eq, Hash, Debug)]
+pub struct ImageCacheKey {
+    source_key: CachedImageSourceKey,
+    gpu_image_flags: ImageRendering,
+}
+
 impl ImageCacheKey {
-    pub fn new(resource: &ImageInner) -> Option<Self> {
+    pub fn new(resource: &ImageInner, gpu_image_flags: Option<ImageRendering>) -> Option<Self> {
         Some(match resource {
             ImageInner::None => return None,
             ImageInner::AbsoluteFilePath(path) => {
                 if path.is_empty() {
                     return None;
                 }
-                Self::Path(path.to_string())
-            }
-            ImageInner::EmbeddedData { data, format: _ } => {
-                Self::EmbeddedData(by_address::ByAddress(data.as_slice()))
+                Self {
+                    source_key: path.to_string().into(),
+                    gpu_image_flags: gpu_image_flags.unwrap_or_default(),
+                }
             }
+            ImageInner::EmbeddedData { data, format: _ } => Self {
+                source_key: by_address::ByAddress(data.as_slice()).into(),
+                gpu_image_flags: gpu_image_flags.unwrap_or_default(),
+            },
             ImageInner::EmbeddedImage { .. } => return None,
         })
     }
@@ -466,7 +484,7 @@ impl ImageCache {
 
     // Try to load the image the given resource points to
     pub(crate) fn load_image_resource(&mut self, resource: &ImageInner) -> Option<Rc<CachedImage>> {
-        ImageCacheKey::new(resource)
+        ImageCacheKey::new(resource, None)
             .and_then(|cache_key| {
                 self.lookup_image_in_cache_or_create(cache_key, || {
                     CachedImage::new_from_resource(resource).map(Rc::new)
