@@ -12,12 +12,12 @@ LICENSE END */
 #[cfg(not(target_arch = "wasm32"))]
 pub fn load_from_path(path: &std::path::Path) -> Result<usvg::Tree, std::io::Error> {
     let svg_data = std::fs::read(path)?;
-    usvg::Tree::from_data(&svg_data, &Default::default())
+    usvg::Tree::from_data(&svg_data, &usvg::Options::default().to_ref())
         .map_err(|e| std::io::Error::new(std::io::ErrorKind::Other, e))
 }
 
 pub fn load_from_data(slice: &[u8]) -> Result<usvg::Tree, usvg::Error> {
-    usvg::Tree::from_data(slice, &Default::default())
+    usvg::Tree::from_data(slice, &usvg::Options::default().to_ref())
 }
 
 pub fn render(
@@ -30,10 +30,10 @@ pub fn render(
     let size = fit.fit_to(tree.svg_node().size.to_screen_size()).ok_or(usvg::Error::InvalidSize)?;
     let mut buffer =
         vec![0u8; size.width() as usize * size.height() as usize * tiny_skia::BYTES_PER_PIXEL];
-    let skya_buffer =
+    let skia_buffer =
         tiny_skia::PixmapMut::from_bytes(buffer.as_mut_slice(), size.width(), size.height())
             .ok_or(usvg::Error::InvalidSize)?;
-    resvg::render(tree, fit, skya_buffer).ok_or(usvg::Error::InvalidSize)?;
+    resvg::render(tree, fit, skia_buffer).ok_or(usvg::Error::InvalidSize)?;
     Ok(image::DynamicImage::ImageRgba8(
         image::RgbaImage::from_raw(size.width(), size.height(), buffer)
             .ok_or(usvg::Error::InvalidSize)?,
